#pragma once

#include <QtQml>

#include "MsXpS/export-import-config.h"

#include "MsXpS/libXpertMassCore/globals.hpp"

namespace MsXpS
{
namespace libXpertMassCore
{
DECLSPEC void registerJsQmlTypes(const char *uri  = "libXpertMassCore",
                                 int versionMajor = 1,
                                 int versionMinor = 0);

DECLSPEC void registerEnumsToQJSEngine(QJSEngine *engine);

DECLSPEC void registerGlobalsToQJSEngine(QJSEngine *engine);


} // namespace libXpertMassCore
} // namespace MsXpS
