
#pragma once

/////////////////////// stdlib includes


/////////////////////// Qt includes
#include <QJSEngine>


/////////////////////// pappsomspp includes


/////////////////////// libXpertMassCore includes


/////////////////////// libXpertMassGui includes


/////////////////////// Local includes
#include "MsXpS/export-import-config.h"

namespace MsXpS
{
namespace libXpertMassGui
{

DECLSPEC void registerEnumsToQJSEngine(QJSEngine *engine);

DECLSPEC void registerGlobalsToQJSEngine(QJSEngine *engine);


} // namespace libXpertMassGui
} // namespace MsXpS
