/* BEGIN software license
 *
 * MsXpertSuite - mass spectrometry software suite
 * -----------------------------------------------
 * Copyright (C) 2009--2026 Filippo Rusconi
 *
 * http://www.msxpertsuite.org
 *
 * This file is part of the MsXpertSuite project.
 *
 * The MsXpertSuite project is the successor of the massXpert project. This
 * project now includes various independent modules:
 *
 * - massXpert, model polymer chemistries and simulate mass spectrometric data;
 * - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * END software license
 */

#pragma once

/////////////////////// stdlib includes


/////////////////////// Qt includes
#include <QList>
#include <QMap>
#include <QFileInfo>


/////////////////////// pappsomspp includes
#include "pappsomspp/core/trace/trace.h"


/////////////////////// Local includes
#include "MsXpS/export-import-config.h"
#include "MsXpS/libXpertMassCore/AbstractLowMassDeconvolver.hpp"

namespace MsXpS
{
namespace libXpertMassCore
{


class DECLSPEC LowMassDeconvolverFullLabelling
  : public AbstractLowMassDeconvolver
{

  public:
  explicit LowMassDeconvolverFullLabelling(Parameters &parameters);

  virtual ~LowMassDeconvolverFullLabelling();

  virtual std::vector<DeconvolutedFeature>
  deconvolute(const pappso::Trace &centroided_mass_spectrum) const;

  protected:
  virtual bool fitIsotopicEnvelope(const pappso::Trace &input_centroids,
                                   std::size_t input_centroid_index,
                                   ChargeSpecificFeature &feature,
                                   std::vector<bool> &globally_used) const;

  virtual std::vector<ChargeSpecificFeature>
  generateChargeSpecificFeatures(const pappso::Trace &input_centroids) const;

  virtual std::vector<IsoMatch>::const_iterator
  getMonoisotopicMatchIterator(const std::vector<IsoMatch> &matches,
                               QString &message) const;

  virtual std::vector<IsoMatch>::const_iterator
  getIsotopologueMatchIterator(const std::vector<IsoMatch> &matches,
                               int isotopologue,
                               QString &message) const;

  pappso::Trace getIsotopicClusterReducedToMaxMatchCount(
    const pappso::Trace &isotopic_cluster) const;
};

} // namespace libXpertMassCore
} // namespace MsXpS
