/* BEGIN software license
 *
 * msXpertSuite - mass spectrometry software suite
 * -----------------------------------------------
 * Copyright (C) 2009--2020 Filippo Rusconi
 *
 * http://www.msxpertsuite.org
 *
 * This file is part of the msXpertSuite project.
 *
 * The msXpertSuite project is the successor of the massXpert project. This
 * project now includes various independent modules:
 *
 * - massXpert, model polymer chemistries and simulate mass spectrometric data;
 * - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * END software license
 */

#pragma once

/////////////////////// StdLib includes


/////////////////////// Qt includes
#include <QWidgetAction>
#include <QCheckBox>
#include <QLabel>

/////////////////////// pappsomspp includes


/////////////////////// Local includes
#include "MsXpS/export-import-config.h"


namespace MsXpS
{
  namespace libXpertMassGui
  {


  class DECLSPEC LabelledCheckBoxAction: public QWidgetAction
  {
    Q_OBJECT

    public:
    LabelledCheckBoxAction(const QString &check_box_text,
                           Qt::CheckState check_state,
                           const QString &action_text,
                           QObject *parent = nullptr);

  virtual ~LabelledCheckBoxAction();

  QCheckBox *getWidget() const;
  QString getText() const;
  Qt::CheckState getCheckState() const;
  bool isChecked() const;

  public slots:
  void checkBoxStateToggled(Qt::CheckState check_state);

  signals:
  void checkBoxStateToggledSignal(Qt::CheckState check_state);

  private:
    QString action_text;
  QCheckBox *mpa_checkBox   = nullptr;
  QLabel *mpa_actionTextLabel = nullptr;
  };


  } // namespace libXpertMassGui
} // namespace MsXpS
