#pragma once

#include <QtCore/QtGlobal>

namespace MsXpS
{

namespace libXpertMassCore
{

#if defined(EXPORT_LIB_SYMBOLS)
// #pragma message("EXPORT_LIB_SYMBOLS is defined: defining DECLSPEC
// Q_DECL_EXPORT")
#define DECLSPEC Q_DECL_EXPORT
#else
// #pragma message("EXPORT_LIB_SYMBOLS is not defined: defining DECLSPEC
// Q_DECL_IMPORT")
#define DECLSPEC Q_DECL_IMPORT
#endif

} // namespace libXpertMassCore

} // namespace MsXpS
