/* BEGIN software license
 *
 * msXpertSuite - mass spectrometry software suite
 * -----------------------------------------------
 * Copyright 2009--2026 by Filippo Rusconi
 *
 * http://www.msxpertsuite.org
 *
 * This file is part of the msXpertSuite project.
 *
 * The msXpertSuite project is the successor of the massXpert project. This
 * project now includes various independent modules:
 *
 * - massXpert, model polymer chemistries and simulate mass spectrometric data;
 * - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * END software license
 */


#pragma once


/////////////////////// Qt includes
#include <QListWidget>


/////////////////////// Local includes
#include "MsXpS/export-import-config.h"

namespace MsXpS
{
namespace libXpertMassGui
{


class DECLSPEC ScriptingHistoryListWidget: public QListWidget
{
  Q_OBJECT
  public:
  ScriptingHistoryListWidget(QWidget *parent = 0);
  ~ScriptingHistoryListWidget();

  QString selectedItemsText();
  void deselectAll();

  private:
  void keyPressEvent(QKeyEvent *event);

  signals:
  void setSelectedItemsTextToScriptInput(const QString &code_text,
                                         bool overwrite);

  public slots:
  void showAll();
  void toggleSelection();

  void hideSelected();
  void removeSelected();
  void removeDuplicates();

  void overwriteSelectedItemsText();
  void insertSelectedItemsText();

  void copySelectedItemsTextToClipboard();

  void selectItemIndices(int startIndex, int endIndex);
};

} // namespace libXpertMassGui
} // namespace MsXpS
